module abstract_inner_product

  use abstract_vector
  implicit none

  private
  public :: inner_product_class


  ! INNER_PRODUCT_CLASS is an abstract type in which we define inner
  ! products. Mathematically, an inner product is defined by a
  ! non-singular, symmetric matrix D: <x,y>=x'Dy . In Fortran, we
  ! define an inner product as a collection of subroutines that return
  ! various products involving the matrix D.

  ! INNER_PRODUCT_CLASS defines the following methods:
  !   dot_v_v           : (deferred) Inner product of two vectors.
  !   dot_g_g           : (deferred) Inner product of two gradients.
  !   D_times_v         : (deferred) Returns Dv, where v is a vector.
  !   D_times_g         : (deferred) Returns Dg, where g is a gradient.
  !   D_inverse_times_v : (deferred) Returns inv(D)*v, where v is a vector.
  !   D_inverse_times_g : (deferred) Returns inv(D)*g, where g is a gradient.

  ! NB: There is probably a way of overloading things so that we don't
  ! have to define everything twice (once for vectors and again for gradients).

  type, abstract :: inner_product_class
    private
    character(len=10), public :: name
  contains
    procedure(dot), deferred :: dot_v_v
    procedure(dot), deferred :: dot_g_g
    procedure(sub), deferred :: D_times_v
    procedure(sub), deferred :: D_times_g
    procedure(sub), deferred :: D_inverse_times_v
    procedure(sub), deferred :: D_inverse_times_g
  end type inner_product_class

  abstract interface
    function dot(this,a,b)
      import :: inner_product_class
      import :: vector_class
      class(inner_product_class), intent(in) :: this
      real :: dot
      class(vector_class), intent(in) :: a,b
    end function
    subroutine sub(this,a)
      import :: inner_product_class
      import :: vector_class
      class(inner_product_class), intent(in) :: this
      class(vector_class), intent(inout) :: a
    end subroutine
  end interface
end module abstract_inner_product
